@@ -28,6 +28,7 @@
 #include  "gdk/gdk.h"
 
 #define	SRL_MAX_CLIENTS			1
+#define SRL_MAX_CHILDREN_CNT 		200	/* depth of child searches */
 #define	SRL_LOG
 /* #define SRL_TIMING */
 
@@ -202,6 +203,7 @@ static Accessible 		*srl_last_context 	= NULL;
 static Accessible		*srl_last_create	= NULL;
 /*FIXME:next varaible should be in SRObject.c file */
 Accessible 			*srl_last_edit		= NULL;
+static gint 			srl_last_row = -1;
 
 extern gboolean sro_set_difference (SRObject *obj, gchar *difference);
 extern gboolean sro_set_name (SRObject *obj, gchar *name);
@@ -593,6 +595,97 @@ srl_acc_has_state (Accessible *acc,
     return rv;
 }
 
+static Accessible*
+srl_get_focused_cell (Accessible *acc_table)
+{
+    Accessible *rv = NULL;
+    AccessibleTable *table;
+    AccessibleComponent *comp;
+    gint rows, rowe, cols, cole;
+
+    srl_assert (acc_table);
+
+    table = Accessible_getTable (acc_table);
+    comp = Accessible_getComponent (acc_table);
+
+    rows = rowe = cols = cole = -1;
+    if (table && comp)
+    {
+	long int x, y, w, h;
+	gint index;
+	Accessible *corner;
+
+	AccessibleComponent_getExtents  (comp, &x, &y, &w, &h,
+						    SPI_COORD_TYPE_WINDOW);
+	/* get the uper-left corner */
+    	corner = AccessibleComponent_getAccessibleAtPoint (comp, x + 1, y + 1,
+					    	    SPI_COORD_TYPE_WINDOW);
+	if (corner)
+	{
+	    index = Accessible_getIndexInParent (corner);
+	    rows = AccessibleTable_getRowAtIndex (table, index);
+	    cols = AccessibleTable_getColumnAtIndex (table, index);
+	    Accessible_unref (corner);
+	}
+	else
+	    rows = cols = 0;
+	
+	/* get the bottom-left corner */
+    	corner = AccessibleComponent_getAccessibleAtPoint (comp, x + 1, y + h - 1,
+					    	    SPI_COORD_TYPE_WINDOW);
+	if (corner)
+	{
+	    index = Accessible_getIndexInParent (corner);
+	    rowe = AccessibleTable_getRowAtIndex (table, index);
+	    Accessible_unref (corner);
+	}
+	else
+	    rowe = AccessibleTable_getNRows (table);
+
+	/* get the upper-right corner */
+    	corner = AccessibleComponent_getAccessibleAtPoint (comp, x + w - 1, y + 1,
+					    	    SPI_COORD_TYPE_WINDOW);
+	if (corner)
+	{
+	    index = Accessible_getIndexInParent (corner);
+	    cole = AccessibleTable_getColumnAtIndex (table, index);
+	    Accessible_unref (corner);
+	}
+	else
+	    cole = AccessibleTable_getNColumns (table);
+	
+    }
+
+    if (rows != -1 && cols != -1)
+    {
+	/* limit our search to at most SRL_MAX_CHILDREN_CNT cells, in */
+	/* case we get pathological start/end row/col (or have a large */
+	/* screen with a tiny spreadsheet with a gazillion showing cells */
+	gint i, j;
+	gint count = 0;
+	for (i = rows; !rv && i <= rowe && count < SRL_MAX_CHILDREN_CNT; i++)
+	    for (j = cols; !rv && j <= cole && count < SRL_MAX_CHILDREN_CNT; j++)
+	    {
+		count++;
+		Accessible *cell = AccessibleTable_getAccessibleAt (table, i, j);
+		if (cell)
+		{
+		    if (srl_acc_has_state (cell, SPI_STATE_FOCUSED))
+			rv = cell;
+		    else
+		        Accessible_unref (cell);
+		}
+	    }
+    }
+
+    if (table)
+	AccessibleTable_unref (table);
+    if (comp)
+	AccessibleComponent_unref (comp);
+
+    return rv;
+}
+
 static gboolean
 srle_acc_has_state (SRLEvent *event,
 		    AccessibleState state)
@@ -670,6 +763,8 @@ srle_set_acc (SRLEvent *event)
 			    Accessible_unref (child);
 		    }
 		}
+		if (!event->acc)
+		    event->acc = srl_get_focused_cell (event->acc_ev->source);
 	    }
 	    
 	    AccessibleSelection_unref (selection);
@@ -898,7 +993,6 @@ srl_set_last_table (Accessible *acc)
 static gboolean
 srl_table_same_line (Accessible *acc)
 {
-    static gint last_row = -1;
     Accessible *parent;
     gboolean rv = FALSE;
 
@@ -912,9 +1006,9 @@ srl_table_same_line (Accessible *acc)
 	{
 	    gint row = AccessibleTable_getRowAtIndex (table, 
 				    Accessible_getIndexInParent (acc));
-	    if (row == last_row && parent == srl_last_table)
+	    if (row == srl_last_row && parent == srl_last_table)
 	        rv = TRUE;
-	    last_row = row;
+	    srl_last_row = row;
 	    AccessibleTable_unref (table);
 	}
 	Accessible_unref (parent);
@@ -962,6 +1056,9 @@ srl_notify_clients_obj (SRLEvent *event,
 	SRObjectRoles role;
 	sro_get_role (obj, &role, SR_INDEX_CONTAINER);
 	
+	if (role == SR_ROLE_TABLE_COLUMNS_HEADER)
+	    srl_last_row = -1;
+
 	if ((role == SR_ROLE_TABLE_LINE || role == SR_ROLE_TABLE_COLUMNS_HEADER) &&
 	    srle_has_type (event, SRL_EVENT_FOCUS))
 	{
@@ -1456,7 +1553,8 @@ srle_change_type (SRLEvent *event)
 {
     srl_assert (event);
     
-    if (srle_has_type (event, SRL_EVENT_ACTIVE_DESCENDANT_CHANGED))
+    if (srle_has_type (event, SRL_EVENT_ACTIVE_DESCENDANT_CHANGED) && 
+	srle_is_for_focused_ancestor_acc (event))
     {
 	event->type = SRL_EVENT_FOCUS;
     }
@@ -1766,6 +1864,7 @@ srl_selection_event_is_reported (SRLEvent *event)
 
 
 static SRLLastInfo last_info_focus, last_info_watched;
+static gboolean srl_set_last_focus (Accessible *acc);
 
 static gboolean
 srl_event_is_reported (SRLEvent *event)
@@ -1849,6 +1948,8 @@ srl_event_is_reported (SRLEvent *event)
 		 srle_acc_has_role (event, SPI_ROLE_SLIDER);
 	else if (srle_has_type (event, SRL_EVENT_VISIBLE_DATA_CHANGED))
 	    rv = TRUE;
+	else if (srle_has_type (event, SRL_EVENT_STATE_SELECTED))
+	    rv = srl_last_table != NULL; /* selection changed while in a table */
     }
 
     if (srle_has_type (event, SRL_EVENT_WINDOW_DEACTIVATE))
@@ -2082,7 +2183,8 @@ srl_init ()
     srl_last_table	= NULL;
     srl_last_context 	= NULL;
     srl_last_create 	= NULL;
-    
+    srl_last_row 	= -1;
+
     srl_log_init ();
     srl_timing_init ();
 
